/*
 * sox.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

typedef sox_effects_chain_t chain_t;
typedef sox_effect_handler_t const handler_t;
typedef sox_signalinfo_t signal_t;

static sox_effect_t *ffsox_effect(sox_effect_handler_t const *handler,
    int n, char **args)
{
  sox_effect_t * effp=NULL;

  if (NULL==handler)
    goto error;

  if (NULL==(effp=sox_create_effect(handler)))
    goto error;

  if (0<n&&SOX_SUCCESS!=sox_effect_options(effp,n,args))
    goto error;

  return effp;
error:
  if (NULL!=effp)
    sox_delete_effect(effp);

  return NULL;
}

static int ffsox_add(chain_t *chain, signal_t *in, signal_t *out,
    handler_t *handler, int n, char **args)
{
  int state=SOX_EOF;
  sox_effect_t *effp=NULL;

  if (NULL==handler||NULL==(effp=ffsox_effect(handler,n,args)))
    goto error;

  if (SOX_SUCCESS!=(state=sox_add_effect(chain,effp,in,out)))
    goto error;

  return SOX_SUCCESS;
error:
  if (NULL!=effp)
    sox_delete_effect(effp);

  return state;
}

static int ffsox_find(chain_t *chain, signal_t *in, signal_t *out,
    char *name, int n, char **args)
{
  return ffsox_add(chain,in,out,sox_find_effect(name),n,args);
}

sox_effects_chain_t *ffsox_chain(audio_t *a)
{
  typedef char buf_t[32];
  enum { N=12 };
  sox_effects_chain_t *chain=NULL;
  decoder_t *dec=a->dec;
  prefs_t *ps=dec->ps;
  sox_signalinfo_t in;
  sox_signalinfo_t out;
  char *args[N],**argp;
  buf_t buf[N],*bufp;
  int dither=1;

  // setup the input signal
  in.rate=a->in.rate;
  in.channels=a->in.nch;
  in.precision=a->in.bps;
  in.length=SOX_UNSPEC;
  in.mult=NULL;

  // create the effects chain.
  if (NULL==(chain=sox_create_effects_chain(NULL,NULL)))
    goto error;

  // add the reader effect.
  argp=args;
  bufp=buf;
  *argp++=(char *)&a->ar;

  if (SOX_SUCCESS!=ffsox_add(chain,&in,&in,ffsox_sox_reader(),argp-args,args))
    goto error;

  // add the rate effect.
  if (a->out.rate!=in.rate) {
    out=in;
    out.rate=a->out.rate;
    argp=args;
    bufp=buf;

    if (0!=ps->sox.rate_q[0])
      *argp++=ps->sox.rate_q;

    sprintf(*bufp,"%.f",out.rate);
    *argp++=*bufp++;

    if (SOX_SUCCESS!=ffsox_find(chain,&in,&out,"rate",argp-args,args))
      goto error;

    dither=1;
  }

  // add the dither effect.
  if (out.precision<32&&dither&&0!=strcmp("off",ps->sox.dither)) {
    out=in;
    out.precision=a->out.bps;
    argp=args;
    bufp=buf;

    if (0==strncmp("-f ",ps->sox.dither,3)) {
      *argp++="-f";
      *argp++=ps->sox.dither+3;
    }
    else if (0!=ps->sox.dither[0])
      *argp++=ps->sox.dither;

    *argp++="-p";
    sprintf(*bufp,"%d",out.precision);
    *argp++=*bufp++;

    if (SOX_SUCCESS!=ffsox_find(chain,&in,&out,"dither",argp-args,args))
      goto error;
  }

  // add the writer effect.
  out=in;
  out.precision=a->out.bps;
  argp=args;
  bufp=buf;
  *argp++=(char *)a;

  if (SOX_SUCCESS!=ffsox_add(chain,&in,&out,ffsox_writer(),argp-args,args))
    goto error;

  return chain;
error:
  if (NULL!=chain)
    sox_delete_effects_chain(chain);

  return NULL;
}
