/*
 * video.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

video_t *video_init(video_t *v, decoder_t *dec, In_Module *mod)
{
  v->dec=dec;
  v->mod=mod;

  if (NULL==(v->frame=avcodec_alloc_frame()))
    goto error;

  avcodec_get_frame_defaults(v->frame);
  timer_init(&v->timer);

  if (NULL==queue_init(&v->q,dec,"Video",dec->ps->video.queue_size))
    goto error;

  if (NULL!=mod&&NULL==screen_init(&v->s,mod))
    goto error;

  return v;
error:
  video_cleanup(v);

  return NULL;
}

video_t *video_cleanup(video_t *v)
{
  screen_cleanup(&v->s);
  queue_cleanup(&v->q);
  timer_cleanup(&v->timer);

  if (NULL!=v->frame)
    av_free(v->frame);

  return v;
}

video_t *video_realloc(video_t *v, track_t *t)
{
  ffi_st_t *vst=t->ffi.vst;
  AVStream *st=vst->st;
  AVCodecContext *dc=vst->dc;
  char *crop;
  double ar1,ar2;

  v->st=vst;
  dc=vst->dc;

  if (NULL!=(crop=track_get_tag(t,"video_crop"))
      &&4==sscanf(crop,"%d:%d:%d:%d",&v->width,&v->height,&v->x,&v->y)) {

    // x
    if (v->width<v->x)
      v->x=v->width;

    // y
    if (v->height<v->y)
      v->y=v->height;

    // width
    if (0==v->width)
      v->width=dc->width;

    if (dc->width<v->x+v->width)
      v->width=dc->width-v->x;

    if (v->width<0)
      v->width=0;

    // height
    if (0==v->height)
      v->height=dc->height;

    if (dc->height<v->y+v->height)
      v->height=dc->height-v->y;

    if (v->height<0)
      v->height=0;

    v->crop=1;
  }
  else {
    v->width=dc->width;
    v->height=dc->height;
    v->x=0;
    v->y=0;
    v->crop=0;
  }

  ar1=st->sample_aspect_ratio.num?av_q2d(st->sample_aspect_ratio)
      :dc->sample_aspect_ratio.num?av_q2d(dc->sample_aspect_ratio)
      :0.0
      ;
  ar2=ar1<=0.0?1.0:ar1;

  screen_open(&v->s,v->width,v->height,ar2);
  avpicture_alloc(&v->pict,dc->pix_fmt,dc->width,dc->height);
  v->time_base=1000.0*av_q2d(st->time_base);

  return v;
#if 0 // {
error:
  return NULL;
#endif // }
}

void video_reset(video_t *v)
{
  prefs_t *ps=v->dec->ps;

  avcodec_get_frame_defaults(v->frame);
  v->timer.state=STATE_RUN;
  v->next_pts=0;
  queue_reset(&v->q,ps->video.queue_size);
}

video_t *video_run(video_t *v)
{
  v->hThread=CreateThread(
    NULL,   // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,      // _In_       SIZE_T dwStackSize,
    video_thread,
            // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    v,      // _In_opt_   LPVOID lpParameter,
    0,      // _In_       DWORD dwCreationFlags,
    NULL    // _Out_opt_  LPDWORD lpThreadId
  );

  return NULL==v->hThread?NULL:v;
}

void video_wait(video_t *v)
{
  if (NULL!=v->hThread) {
    WaitForSingleObject(v->hThread,INFINITE);
    CloseHandle(v->hThread);
    v->hThread=NULL;
  }
}

void video_stop(video_t *v)
{
  screen_done(&v->s);
  avpicture_free(&v->pict);
  v->st=NULL;
}

void video_kill(video_t *v)
{
  timer_kill(&v->timer);
}

#define video_drift(v) \
  (round((v)->time_base*(v)->frame->pts-decoder_get_time((v)->dec,(v)->mod)))

void video_display_frame(video_t *v)
{
  decoder_t *dec=v->dec;
  prefs_t *ps=dec->ps;
  ffi_st_t *vst=v->st;
  AVCodecContext *dc=vst->dc;
  AVFrame *frame=v->frame;
  AVPicture crop;
  AVPicture *pict;
  double drift;

  // implies in_pix_fmt == out_pix_fmt
  if (!v->frame->interlaced_frame)
    pict=(AVPicture *)frame;
  else {
    avpicture_deinterlace(&v->pict,(AVPicture *)frame,dc->pix_fmt,
        dc->width,dc->height);
    pict=&v->pict;
  }

  if (v->crop&&(0<v->x||0<v->y)) {
    av_picture_crop(&crop,pict,dc->pix_fmt,v->y,v->x);
    pict=&crop;
  }

  frame->pts=av_frame_get_best_effort_timestamp(frame);
  drift=video_drift(v);

  if (ps->video.max_drift<=drift) {
#if defined (DEBUG)
    fprintf(stderr,"DRIFT 1: %.0f\n",drift);
#endif
    frame->pts=v->next_pts;
    drift=video_drift(v);
  }

  if (ps->video.min_drift<=drift&&drift<ps->video.max_drift) {
    if (0.0<drift) {
#if defined (DEBUG)
      //fprintf(stderr,"DRIFT 2: %.0f\n",drift);
#endif
      timer_sleep(&v->timer,drift);
    }
#if defined (DEBUG)
	else {
      //fprintf(stderr,"DRIFT 3: %.0f\n",drift);
    }
#endif

    if (STATE_KILL!=v->timer.state)
      screen_draw(&v->s,pict);

	v->next_pts=frame->pts+frame->pkt_duration;
  }
#if defined (DEBUG)
  else {
    if (STATE_KILL!=v->timer.state)
      screen_draw(&v->s,pict);
    fprintf(stderr,"DRIFT 4: %.0f\n",drift);
  }
#endif

  avcodec_get_frame_defaults(frame);
}

void video_display(video_t *v, AVPacket *pkt, int *got_frame)
{
  ffi_st_t *vst=v->st;
  AVCodecContext *dc=vst->dc;

  if (avcodec_decode_video2(dc,v->frame,got_frame,pkt)<0)
    goto cleanup;

  if (*got_frame)
    video_display_frame(v);
cleanup:
  return;
}

DWORD WINAPI video_thread(LPVOID lpParam)
{
  video_t *v=lpParam;
  AVPacket pkt;
  int got_frame;

  while (0==queue_read(&v->q,&pkt)) {
    video_display(v,&pkt,&got_frame);
    av_free_packet(&pkt);
  }

  if (STATE_KILL!=v->timer.state) {
    // flush cached frames
    pkt.data=NULL;
    pkt.size=0;

    do {
      video_display(v,&pkt,&got_frame);
    } while (got_frame);
  }

  return 0;
}
